(ns com.aurellem.run.bootstrap-1
  (:use (com.aurellem.gb saves gb-driver util
                         items vbm characters money))
  (:use (com.aurellem.run util title save-corruption bootstrap-0))
  (:use (com.aurellem.exp item-bridge))
  (:import [com.aurellem.gb.gb_driver SaveState]))

(defn pc-item-writer-program
  []
  (let [limit 201
        [target-high target-low] (disect-bytes-2 0xD162)]
    (flatten
     [[0x00  ;; (item-hack) set increment stack pointer no-op
       0x1E  ;; load limit into E
       limit
       0x3F  ;; (item-hack) set carry flag no-op

       ;; load 2 into C.
       0x0E   ;; C == 1 means input-first nybble
       0x04   ;; C == 0 means input-second nybble

       0x21 ;; load target into HL
       target-low
       target-high
       0x37 ;; (item-hack) set carry flag no-op

       0x00 ;; (item-hack) no-op
       0x37 ;; (item-hack) set carry flag no-op
       
       0x00 ;; (item-hack) no-op
       0xF3 ;; disable interrupts
       ;; Input Section

       0x3E ;; load 0x20 into A, to measure buttons
       0x10 

       0x00 ;; (item-hack) no-op
       0xE0 ;; load A into [FF00]
       0x00

       0xF0 ;; load 0xFF00 into A to get
       0x00 ;; button presses
       
       0xE6
       0x0F ;; select bottom four bits of A
       0x37 ;; (item-hack) set carry flag no-op

       0x00 ;; (item-hack) no-op
       0xB8 ;; see if input is different (CP A B)

       0x00 ;; (item-hack) (INC SP)
       0x28 ;; repeat above steps if input is not different
       ;; (jump relative backwards if B != A)
       0xED ;; (literal -19) (item-hack) -19 == egg bomb (TM37)

       0x47 ;; load A into B
       
       0x0D ;; dec C
       0x37 ;; (item-hack) set-carry flag
       ;; branch based on C:
       0x20 ;; JR NZ
       23 ;; skip "input second nybble" and "jump to target" below
       
       ;; input second nybble

       0x0C ;; inc C
       0x0C ;; inc C

       0x00 ;; (item-hack) no-op
       0xE6 ;; select bottom bits
       0x0F
       0x37 ;; (item-hack) set-carry flag no-op

       0x00 ;; (item-hack) no-op
       0xB2 ;; (OR A D) -> A

       0x22 ;; (do (A -> (HL)) (INC HL))

       0x1D ;; (DEC E)

       0x00 ;; (item-hack) 
       0x20 ;; jump back to input section if not done
       0xDA ;; literal -36 == TM 18 (counter)
       0x01 ;; (item-hack) set BC to literal (no-op)

       ;; jump to target
       0x00  ;; (item-hack) these two bytes can be anything.
       0x01 

       0x00   ;; (item-hack) no-op
       0xBF   ;; (CP A A) ensures Z
       
       0xCA   ;; (item-hack) jump if Z
       target-low
       target-high
       0x01   ;; (item-hack) will never be reached.
       
       ;; input first nybble
       0x00
       0xCB
       0x37  ;; swap nybbles on A

       0x57  ;; A -> D

       0x37  ;; (item-hack) set carry flag no-op
       0x18  ;; relative jump backwards
       0xCD  ;; literal -51 == TM05; go back to input section
       0x01  ;; (item-hack) will never reach this instruction

       ]
      (repeat 8 [0x00 0x01])

      [;; jump to actual program
       0x00
       0x37  ;; (item-hack) set carry flag no-op

       0x2E  ;; 0x3A -> L
       0x3A


       0x00  ;; (item-hack) no-op
       0x26  ;; 0xD5 -> L
       0xD5  
       0x01  ;; (item-hack) set-carry BC

       0x00  ;; (item-hack) these can be anything
       0x01  

       0x00
       0xE9 ;; jump to (HL)
       ]])))

(defn view-desired-item-layout []
  (clojure.pprint/pprint
   (raw-inventory->inventory (pc-item-writer-program))))

(defn pc-item-writer-state []
    (-> (read-state "bootstrap-init")
      (set-memory pc-item-list-start 50)
      (set-memory-range
       map-function-address-start
       [0x8B 0xD5])
      (set-memory-range
       (inc pc-item-list-start)
       (pc-item-writer-program))))

(defn test-pc-item-writer []
  (let [orig (read-state "pc-item-writer")]
    (-> orig
    (print-listing 0xD162 (+ 0xD162 20))
    (run-moves (reduce concat
                (repeat 10 [[:a :b :start :select] []])))
    ((fn [_] (println "===========") _))
    (print-listing 0xD162 (+ 0xD162 20)))))

(defn close-all-menus [[moves state :as script]]
  (loop [s script]
    (let [depth (current-depth (second (do-nothing 50 s)))]
      (println "depth" depth)
      (if (= depth 1)
        s
        (recur (close-menu s))))))

(defn-memo name-rival
  ([] (name-rival (to-rival-name)))
  ([script]
     (->> script
          (first-difference [] [:a] AF)
          (first-difference [] [:r] DE)
          (play-moves
           [[]
            [] [] [:r] [] [:d] [:a]           ;; L
            [:r] [] [:r] [] [:r] [] [:r] [] 
            [:r] [] [:d] [] [:d] [:a]         ;; [PK]
            [:d] [] [:r] [:a]
            ]))))

(defn-memo to-room-pc
  ([] (to-room-pc (name-rival)))
  ([script]
     (->> script
          finish-title
          (walk [← ← ↑ ← ↑ ↑ ↑]))))

(defn-memo bootstrap-corrupt-save
  ([] (bootstrap-corrupt-save (to-room-pc)))
  ([script]
   (->> script
        (do-save-corruption 2)
        (corrupt-item-list 0)
        close-all-menus)))

(defn-memo begin-initial-deposits
  ([] (begin-initial-deposits
       (bootstrap-corrupt-save)))
  ([script]
     (->> script
          (first-difference [] [:a] AF)
          (scroll-text)
          (set-cursor 1)
          select-menu-entry)))

(defn wait-for-quantity
  [[moves state :as script]]
  (if (not= (item-quantity-selected state) 1)
    (repeat-until-different [] item-quantity-selected script)
    script))

(defn wait-for-cursor
  [[moves state :as script]]
  (if (not= (list-offset state) 0)
    (repeat-until-different [] list-offset script)
    script))

(defn deposit-held-item [n quantity [moves state :as script]]
  (let [total-quantity (second (nth-item state n))]
    (println "total-quantity" total-quantity)
    (->> script
         (set-cursor n)
         (select-menu-entry 1)
         (wait-for-quantity)
         (set-quantity total-quantity quantity)
         (delayed-difference [] [:a] 100 #(search-string % "stored"))
         (scroll-text))))

(defn sell-held-item [n quantity [moves state :as script]]
  (let [total-quantity (second (nth-item state n))]
    (->> script
         (wait-for-cursor)  ;; when selling, the cursor always
         (set-cursor n)     ;; returns to the top of the list.
         (select-menu-entry 1)
         (wait-for-quantity)
         (set-quantity total-quantity quantity)
         (delayed-difference [] [:a] 100 current-depth)
         (play-moves (repeat 20 [:b]))
         (delayed-difference [] [:a] 100 #(search-string % "What"))
         )))

(defn widthdraw-pc-item [n quantity [moves state :as script]]
  (let [total-quantity (second (nth-pc-item state n))]
    (->> script
         (set-cursor n)
         (select-menu-entry 1)
         (wait-for-quantity)
         (set-quantity total-quantity quantity)
         (delayed-difference [] [:a] 100 #(search-string % "Withdrew"))
         (scroll-text))))

(defn toss-held-item [n quantity [moves state :as script]]
  (let [total-quantity (second (nth-item state n))]
    (->> script
         (set-cursor n)
         (select-menu-entry 1)
         (set-cursor-relative 1)
         (select-menu-entry -1)
         (wait-for-quantity)
         (set-quantity total-quantity quantity)
         (play-moves [[:a]])
         (scroll-text)
         (delayed-difference [] [:a] 100 #(search-string % "Threw"))
         (scroll-text)
         )))

(defn buy-item [n quantity [moves state :as script]]
  (->> script
       (set-cursor n)
       (purchase-item quantity)))


(def desired-zero-quantities
  (map second (filter (comp (partial = 0) first)
                      (partition 2 (pc-item-writer-program)))))
  
(defn-memo initial-deposits
  ([] (initial-deposits (begin-initial-deposits)))
  ([script]
     (->> script
          (deposit-held-item 0 0x1)
          ((fn [script]
             (reduce 
              (fn [script item] (deposit-held-item item 0xFF script))
              script
              (range 3 (+ 13 3)))))
          close-all-menus)))


(defn-memo prepare-celadon-warp
  ([] (prepare-celadon-warp (initial-deposits)))
  ([script]
     (->> script
          (activate-start-menu)
          (set-cursor-relative 1)
          (select-menu-entry)
          (toss-held-item 35 0xFA)
          (close-all-menus))))


;;0 -- 256
;;1 -- 254
;;2 -- 254
;;3 -- 255

(defn-memo restore-items
  ([] (restore-items (prepare-celadon-warp)))
  ([script]
     (->> script
          (first-difference [] [:a] AF)
          (scroll-text)
          (select-menu-entry)
          (widthdraw-pc-item 0 1)
          ;;(widthdraw-pc-item 0 99)
          ;;(widthdraw-pc-item 1 1)
          (widthdraw-pc-item 13 255)
          (close-all-menus))))

(defn-memo to-celadon
  ([] (to-celadon (restore-items)))
  ([script]
     (->> script
          (walk [→ → → → → → → ↑
                 ↓ ↓ ↓ ↓ ↓ ← ← ← ←
                 ↓ ↓]))))


;; celadon store inventory

;; Floor 2
;;=====================================
;; Great Ball        TM32 (double-team)
;; Super Potion      TM33 (reflect)
;; Revive            TM02 (razor-wind)
;; Super Repel       TM07 (horn-drill)
;; Antidote          TM37 (egg-bomb)
;; Burn Heal         TM01 (mega-punch)
;; Ice Heal          TM05 (mega-kick)
;; Awakening         TM09 (take-down)
;; Parlyz Heal       TM17 (submission)


;; Floor 3
;;=====================================       
;; TM18 (counter)


;; Floor 4
;;=====================================
;; Poke Doll
;; Fire Stone
;; Thunder Stone
;; Water Stone
;; Leaf Stone

;; Floor 5
;;=====================================
;; X Accuracy         HP UP
;; Guard Spec.        Protein
;; Dire Hit           Iron
;; X Attack           Carbos
;; X Defend           Calcium
;; X Speed            
;; X Special

;; Roof
;;=====================================
;; Fresh Water        TM13 (ice-beam)
;; Soda Pop           TM48 (rock-slide)
;; Lemonade :)        TM49 (tri-attack)


(defn-memo go-to-floor-two
  ([] (go-to-floor-two (to-celadon)))
  ([script]
     (->> script
          (walk [↑ → → → → → → → → → → →
                 ↑ ↑ ↑ ↑ ↑ ↑
                 ← ← ← ← 
                 ↓ ↓ ↓
                 ← ←])
          (first-difference [] ↑ AF))))

(defn talk
  "Assumes that you are facing something that initiates text and
   causes it to do so."
  [script]
  (->> script
       (delayed-difference [] [:a] 100
                           #(aget (memory %) text-address))))

(defn-memo get-money-floor-two
  ([] (get-money-floor-two (go-to-floor-two)))
  ([script]
     (->> script
          talk
          (set-cursor 1)
          (select-menu-entry)
          (sell-held-item 0 1)
          (sell-held-item 0 1)
          (close-menu))))

(defn-memo floor-two-TMs
  ([] (floor-two-TMs (get-money-floor-two)))
  ([script]
     (->> script
          (wait-for-cursor)
          (select-menu-entry)
          (buy-item 2 98)  ;; TM02 (razor-wind)
          (buy-item 4 71)  ;; TM37 (doubleteam)
          (buy-item 5 63)  ;; TM01 (mega-punch)
          (buy-item 6 1)   ;; TM05 (mega-kick)
          (buy-item 7 56)  ;; TM09 (take-down)
          (close-menu))))

(defn end-shop-conversation
  [script]
  (->> script
       (wait-until scroll-text [:b])
       (play-moves [[] [:b]])
       close-menu))

(defn-memo floor-two-more-money
  ([] (floor-two-more-money (floor-two-TMs)))
  ([script]
     (->> script
          (wait-for-cursor)
          (set-cursor 1)
          (select-menu-entry)
          (sell-held-item 0 1)
          (sell-held-item 0 1)
          close-menu
          end-shop-conversation)))

(defn turn [direction script]
  (->> script
       (first-difference [] direction AF)))

(defn-memo floor-two-items
  ([] (floor-two-items (floor-two-more-money)))
  ([script]
     (->> script
          (walk [←])
          (turn ↑)
          talk
          select-menu-entry
          (buy-item 5 12)  ;; burn heal
          (buy-item 6 55)  ;; ice heal
          (buy-item 7 4)   ;; awakening
          (buy-item 8 99)  ;; parlyz heal
          (buy-item 8 55)  ;; parlyz heal
          close-menu
          end-shop-conversation)))

(defn-memo go-to-floor-three
  ([] (go-to-floor-three (floor-two-items)))
  ([script]
     (->> script
          (walk [→ → → → → → → → → → ↑ ↑ ↑
                   → ↑]))))
(defn-memo get-TM18
  ([] (get-TM18 (go-to-floor-three)))
  ([script]
     (->> script
          (walk [↓ ↓])
          talk
          (scroll-text 3)
          end-text)))

(defn-memo go-to-floor-four
  ([] (go-to-floor-four (get-TM18)))
  ([script]
     (->> script
          (walk [← ← ← ← ↑ ↑
                 ↓ ← ← ↓ ↓ ↓
                 ← ← ← ← ←])
          (turn ↓))))

(defn-memo floor-four-items
  ([] (floor-four-items (go-to-floor-four)))
  ([script]
     (->> script
          talk
          select-menu-entry
          (buy-item 1 23)  ;; Fire Stone
          (buy-item 2 98)  ;; Thunder Stone
          (buy-item 3 29)  ;; Water Stone
          close-menu
          end-shop-conversation)))

(defn-memo go-to-floor-five
  ([] (go-to-floor-five (floor-four-items)))
  ([script]
     (->> script
          (walk [→ → → → → →
                 ↑ ↑ ↑
                 → → → → → ↑ ;; leave floor four
                 ↓ ← ← ← ← ← ← ← ←
                 ↓ ↓ ↓ ← ← ← ]);; go to five's clerk
          (turn ↑))))
                
(defn-memo floor-five-items
  ([] (floor-five-items (go-to-floor-five)))
  ([script]
     (->> script
          talk
          select-menu-entry
          (buy-item 0 58)   ;; X-Accuracy
          (buy-item 1 99)   ;; Guard Spec.
          (buy-item 1 24)   ;; Guard Spec.
          close-menu
          end-shop-conversation)))

(defn-memo go-to-roof
  ([] (go-to-roof (floor-five-items)))
  ([script]
     (->> script
          (walk [→ → → → ↑ ↑ ↑ → → → ↑ ;; leave floor five
                 ↓ ← ← ←]) ;; walk to vending machine
          (turn ↑))))

(defn buy-drink
  "Assumes you're in front of the vending machine. Buys the indicated
   drink."
  [n script]
  (->> script
       (do-nothing 20)
       (play-moves [[:a][:a]])
       scroll-text
       (wait-for-cursor)
       (set-cursor n)
       select-menu-entry
       close-menu))

(defn-memo roof-drinks
  ([] (roof-drinks (go-to-roof)))
  ([script]
     (->> script
          (buy-drink 0) ;; fresh water (for TM13)
          ;; buy 16 lemonades
          ;; LEMONADE is the best item <3  :)
          (multiple-times 16 (partial buy-drink 2)))))

(defn get-TM13
  ([] (get-TM13 (roof-drinks)))
  ([script]
     (->> script
          (walk [← ← ← ← ← ← ↓])
          (play-moves [[][:a][:a][]])
          (scroll-text 3)
          select-menu-entry
          select-menu-entry
          (scroll-text 6)
          close-menu)))
          
